/**
  *  \file client/dialogs/selmgr.cc
  *  \brief Selection Manager Dialog
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/selmgr.h"
#include "client/widgets/selinfo.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "game/global.h"
#include "game/sel.h"
#include "game/univ.h"
#include "gfx/palette.h"
#include "int/interror.h"
#include "int/lex.h"
#include "int/selexpr.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/widgets.h"
#include "ui/widgets/focusgroup.h"
#include "ui/widgets/focusiterator.h"
#include "ui/window.h"
#include "util/nls.h"

/** Selection Manager Dialog. */
class WSelectionManager : public UIWindow {
    ObjectHolder h;

    void init();
    void updateUI();
 public:
    WSelectionManager();
    ~WSelectionManager();

    int getCurrentLayer();
    bool handleEvent(const UIEvent& event, bool second_pass);
    void onEnter();
};

/** Constructor. */
WSelectionManager::WSelectionManager()
    : UIWindow(_("Selection Manager"), BLUE_WINDOW, COLOR_GRAY)
{
    init();
}

/** Destructor. */
WSelectionManager::~WSelectionManager()
{ }

/** Create the dialog. */
void
WSelectionManager::init()
{
    // VBox
    //   HBox
    //     VBox
    //       UIFocusGroup { WSelectionLayerInfo Layer A }
    //       ...
    //       UIFocusGroup { WSelectionLayerInfo Layer H }
    //     VBox
    //       UIButton "D-Clear"
    //       UIButton "I-Invert"
    //       UIButton "C-Copy"
    //      (UIButton "S-Save"
    //       UIButton "R-Load"
    //       UIButton "M-Merge")
    //       UISpacer
    //   HBox
    //     UIButton "Enter-Select"
    //     UIButton "ESC-Close"
    //     UISpacer
    //    (UIButton "H")

    setLayout(UIVBoxLayout::instance5);

    UIGroup& group1  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group11 = h.add(new UIGroup(UIVBoxLayout::instance0));
    UIGroup& group12 = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group2  = h.add(new UIGroup(UIHBoxLayout::instance5));

    for (int i = 0; i < GMultiSelection::NUM_SELECTION_LAYERS; ++i) {
        UIFocusGroup& fg = h.add(new UIFocusGroup(UIHBoxLayout::instance5, 1000 + i));
        fg.add(h.add(new WSelectionLayerInfo(2000 + i, i)));
        group11.add(fg);
        if (i == GMultiSelection::getCurrentSelectionLayer())
            fg.setFocusedWidget();
    }
    group1.add(group11);

    group12.add(h.add(new UIButton(_("D - Clear"),  'd')).addKey());
    group12.add(h.add(new UIButton(_("I - Invert"), 'i')).addKey());
    group12.add(h.add(new UIButton(_("C - Copy"),   'c')).addKey());
    group12.add(h.add(new UISpacer()));
    group1.add(group12);
    add(group1);

    group2.add(h.add(new UIButton(cm_Enter, _("Enter - Select"), SDLK_RETURN)).addCall(this, &WSelectionManager::onEnter));
    group2.add(h.add(new UIButton(cm_Escape, _("ESC - Close"), SDLK_ESCAPE)).addStop());
    group2.add(h.add(new UISpacer()));
    add(group2);

    add(h.add(new UIFocusIterator(1000, 1000 + GMultiSelection::NUM_SELECTION_LAYERS-1,
                                  UIFocusIterator::fi_Vert | UIFocusIterator::fi_Wrap | UIFocusIterator::fi_Tab)));

    pack();
}

/** Update display after possible changes to the universe. */
void
WSelectionManager::updateUI()
{
    getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
}

/** Get currently focused selection layer. */
int
WSelectionManager::getCurrentLayer()
{
    // Locate a widget that has an Id in [1000, 1000+NUM_SELECTION_LAYERS).
    // This will be the selected focus group containing the current layer.
    UIBaseWidget* w = getFocusedWidget();
    while (w != 0 && w != this) {
        if (w->getId() >= 1000 && w->getId() < 1000 + GMultiSelection::NUM_SELECTION_LAYERS)
            return w->getId() - 1000;
        w = w->getParent();
    }
    return 0;
}

bool
WSelectionManager::handleEvent(const UIEvent& event, bool second_pass)
{
    static const char null_code[] = { GMultiSelection::opZero, 0 };
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'c':
            // Copy
            {
                string_t expr;
                string_t compiledExpr;
                if (UIInputLine::run(expr, 2000, _("Selection Manager"), _("Enter layer/expression to copy from:"), 40, 0)) {
                    try {
                        IntTokenizer tok(expr);
                        compileSelectionExpression(tok, compiledExpr);
                        if (tok.getCurrentToken() != tok.tEnd)
                            throw IntError::garbageAtEnd(true);
                        GMultiSelection::executeSelectionExpression(getCurrentLayer(), compiledExpr);
                        updateUI();
                    }
                    catch (IntError& e) {
                        messageBox(format(_("Invalid selection expression: %s"), e.what()), _("Selection Manager"));
                    }
                }
            }
            return true;
         case 'd':
            // Delete
            if (UIMessageBox(_("Do you want to clear this selection layer?"), _("Selection Manager")).doYesNoDialog()) {
                GMultiSelection::executeSelectionExpression(getCurrentLayer(), null_code);
                updateUI();
            }
            return true;
         case 'd' + ss_Ctrl:
            // Delete all
            if (UIMessageBox(_("Do you want to clear all selection layers?"), _("Selection Manager")).doYesNoDialog()) {
                for (int i = 0; i < GMultiSelection::NUM_SELECTION_LAYERS; ++i)
                    GMultiSelection::executeSelectionExpression(i, null_code);
                updateUI();
            }
            return true;
         case 'i':
            // Invert
            {
                int c = getCurrentLayer();
                char code[] = { GMultiSelection::opFirstLayer + c, GMultiSelection::opNot, 0 };
                GMultiSelection::executeSelectionExpression(c, code);
                updateUI();
            }
            return true;
         case 'i' + ss_Ctrl:
            // Invert all
            for (int i = 0; i < GMultiSelection::NUM_SELECTION_LAYERS; ++i) {
                char code[] = { GMultiSelection::opFirstLayer + i, GMultiSelection::opNot, 0 };
                GMultiSelection::executeSelectionExpression(i, code);
            }
            updateUI();
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

void
WSelectionManager::onEnter()
{
    GMultiSelection::setCurrentSelectionLayer(getCurrentLayer());
    stop(0);
}

/** Display the "Selection Manager" dialog.
    This dialog allows the user to manipulate the current selections.
    It is invoked by the command 'UI.SelectionManager'. */
void
doSelectionManager()
{
    // Update multi-selection view from game
    GMultiSelection::getSelectionLayer(GMultiSelection::getCurrentSelectionLayer())
        .copyFrom(getDisplayedTurn().getCurrentUniverse());

    // Do dialog
    WSelectionManager mgr;
    ui_root->centerWidget(mgr);
    ui_root->runModal(mgr);
}
