/**
  *  \file int/process.cc
  *  \brief Interpreter: Process Management
  *
  *  We distinguish between two kinds of processes, regular processes
  *  and temporary processes.
  *
  *  Regular processes are added to the process list normally, and can
  *  do everything they might want to do, including suspending. Such
  *  processes are created using createProcess. The user then sets up
  *  the process and runs it using runRunnableProcesses,
  *  killTerminatedProcesses
  *
  *  Temporary processes are used whenever the PCC2 core needs to
  *  evaluate an expression, and wishes immediate feedback. Such
  *  processes are not allowed to suspend; when they try to, they
  *  receive a (catchable) error. Such processes are created by simply
  *  creating a IntExecutionContext object on the stack, setting it
  *  up, and and giving it to runTemporaryProcess. If that returns
  *  true, the process has produced its result as requested, otherwise
  *  it didn't. The process never appears in the process list, and is
  *  killed automatically when the IntExecutionContext object goes out
  *  of scope. This is the equivalent to PCC 1.x's simple process-less
  *  expression evaluation. PCC2 doesn't support process-less
  *  evaluation because it supports user-defined functions.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <list>
#include "int/process.h"
#include "int/exec.h"
#include "util/misc.h"

/*********************** ProcessList utility class ***********************/

namespace {
    class ProcessList : public std::list<IntExecutionContext*> {
     public:
        ~ProcessList();
    };
}

ProcessList::~ProcessList()
{
    safeClear(*this);
}


/******************************* Public API ******************************/

static ProcessList process_list;

/** Create a regular process. The process is created in state psRunnable.
    You have to fill it (by pushing some frames), and then run it by
    runRunnableProcesses().
    \param name Name of process
    \return Reference to blank process object */
IntExecutionContext&
createProcess(string_t name)
{
    return *safeAppend(process_list, new IntExecutionContext(name));
}

/** Run runnable processes. Runs all processes that are ready to run
    (psRunnable). */
void
runRunnableProcesses()
{
    for (ProcessList::iterator i = process_list.begin(); i != process_list.end(); ++i) {
        if ((*i)->getState() == IntExecutionContext::psRunnable) {
            (*i)->run();
            if ((*i)->getState() == IntExecutionContext::psRunnable) {
                // If the process is still runnable, it must be because it performed
                // some UI operation. Exit, and have the caller execute the required
                // operation and call us again.
            }
        }
    }
}

/** Kill terminated processes. Terminates all processes that have exited
    (psTerminated, psError, psEnded). */
void
killTerminatedProcesses()
{
    ProcessList::iterator i = process_list.begin();
    while (i != process_list.end()) {
        switch ((*i)->getState()) {
         case IntExecutionContext::psTerminated:
         case IntExecutionContext::psError:
         case IntExecutionContext::psEnded:
            delete *i;
            i = process_list.erase(i);
            break;
         default:
            ++i;
            break;
        }
    }
}

/** Check whether there's any runnable process. */
bool
hasAnyRunnableProcess()
{
    for (ProcessList::iterator i = process_list.begin(); i != process_list.end(); ++i) {
        if ((*i)->getState() == IntExecutionContext::psRunnable) {
            return true;
        }
    }
    return false;
}

/** Run temporary process. Runs the given process, which is a stack
    object of the caller (not created with createProcess). The process
    is not permitted to suspend.

    \retval true  The process has run till its end and finished normally.
                  If it contained an expression, that expression's result
                  is available on the stack (exec.getResult()).
    \retval false The process has terminated abnormally. This happens when
                  the process received an error, but also when it
                  executed an 'End' statement. In this case, it has not
                  produced a result. */
bool
runTemporaryProcess(IntExecutionContext& exec)
{
    while (1) {
        exec.run();

        /* Do not terminate */
        if (exec.getState() == IntExecutionContext::psSuspended) {
            exec.handleException("Cannot suspend temporary process", string_t());
        }

        /* Process status */
        if (exec.getState() == IntExecutionContext::psRunnable) {
            /* It wants to run again */
        } else if (exec.getState() == IntExecutionContext::psEnded) {
            return true;
        } else {
            // FIXME what to do with UI-wait processes?
            // psError, psTerminated, ...
            return false;
        }
    }
}
